#include <WiFi.h>
#include <Firebase_ESP_Client.h>
#include "time.h"  // Native ESP32 library for NTP server synchronization
#include "DHT.h"

#include "addons/TokenHelper.h"  // Token generation process info
#include "addons/RTDBHelper.h"   // RTDB payload printing info and other helper functions

#include "esp_camera.h"
#include "soc/soc.h"           // Disable brownout problems
#include "soc/rtc_cntl_reg.h"  // Disable brownout problems
#include "driver/rtc_io.h"
#include <SPIFFS.h>
#include <FS.h>


#define BUILTIN_RED_LED 33  // Built-in red LED

#define DHT_PIN 2           // Digital pin connected to the DHT22 sensor
#define DHT_TYPE DHT22      // DHT sensor type
#define DHT_TIMEOUT 5000    // Maximum sampling rate for DHT22 is 2 seconds
unsigned long last_read;    // Last time the DHT22 values were read
DHT dht(DHT_PIN, DHT_TYPE);

#define HOUR_LONG 3600000   // 3600 seconds
unsigned long last_hourly;  // Last time an hourly update took place
float nr_samples;           // Number of readings in the last hour
float temp_med_hour;        // Medium temperature in the last hour
float pres_med_hour;        // Medium pressure in the last hour
byte current_day_5;         // Current day in the last 5
byte current_day_30;        // Current day in the last 30
 
// WiFi network credentials
#define WIFI_SSID "UT-TP-Link_5FDB06"
#define WIFI_PASSWORD "utmultul"

// Firebase project API Key
#define API_KEY "AIzaSyD_szQR21HIyhhIONSyZDv4pbCXx4lpjsA"
 
// RTDB URLefine the RTDB URL
#define DATABASE_URL "https://smart-camera-live-feed-default-rtdb.europe-west1.firebasedatabase.app"

// Firebase storage bucket ID
#define STORAGE_BUCKET_ID "gs://smart-camera-live-feed.appspot.com"

String tempPath = "/temperature";
String humidityPath = "/humidity";
String heatPath = "/heat_index";
String live = "/live";
String last5 = "/last5";
String last30 = "/last30";

// Camera credentials
#define USER_EMAIL "cam1@cameras.com"
#define USER_PASSWORD "KaM-{1}-"
 
FirebaseData fbdo;  // Firebase Data object
FirebaseAuth auth;
FirebaseConfig configF;
bool signupOK = false;

String uid;  // Firebase user ID
String databasePath;  // Main database path
String readingsPath;  // Camera sensor readings node

FirebaseJson readings_json, live_json, last5_json, last30_json;


// Data for date and time
const char* ntpServer = "pool.ntp.org";
const long gmtOffset_sec = 7200;      // UTC +2.00: 2 * 60 * 60
const int daylightOffset_sec = 3600;  // Observes daylight saving time


// Data for camera
// OV2640 camera module pins (CAMERA_MODEL_AI_THINKER)
#define PWDN_GPIO_NUM     32
#define RESET_GPIO_NUM    -1
#define XCLK_GPIO_NUM      0
#define SIOD_GPIO_NUM     26
#define SIOC_GPIO_NUM     27
#define Y9_GPIO_NUM       35
#define Y8_GPIO_NUM       34
#define Y7_GPIO_NUM       39
#define Y6_GPIO_NUM       36
#define Y5_GPIO_NUM       21
#define Y4_GPIO_NUM       19
#define Y3_GPIO_NUM       18
#define Y2_GPIO_NUM        5
#define VSYNC_GPIO_NUM    25
#define HREF_GPIO_NUM     23
#define PCLK_GPIO_NUM     22

#define FILE_PHOTO "/live/cam1.jpg"  // Photo File Name to save in SPIFFS


byte get_local_time() {
  struct tm timeinfo;
  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time");
    return 255;
  }
  Serial.println(&timeinfo, "%A, %B %d %Y %H:%M:%S");
  return timeinfo.tm_hour;
}

// Initialize WiFi
void init_WiFi() {
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
  Serial.print("Connecting to WiFi ..");
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print('.');
    delay(1000);
  }
  Serial.println(WiFi.localIP());
  Serial.println();

  digitalWrite(BUILTIN_RED_LED, LOW);  // turn on LED
}

void init_Firebase() {
  configF.api_key = API_KEY;  // Assign the api key
  configF.database_url = DATABASE_URL;  // Assign the RTDB URL

  // Assign the user sign in credentials
  auth.user.email = USER_EMAIL;
  auth.user.password = USER_PASSWORD;

  Firebase.reconnectWiFi(true);
  fbdo.setResponseSize(4096);

  if (Firebase.signUp(&configF, &auth, "", "")){
    Serial.println("Signup ok");
    signupOK = true;
  }
  else{
    Serial.printf("%s\n", configF.signer.signupError.message.c_str());
  }
 
  // Assign the callback function for the long running token generation task
  configF.token_status_callback = tokenStatusCallback;  //see addons/TokenHelper.h
 
  // Assign the maximum retry of token generation
  configF.max_token_generation_retry = 5;
 
  // Initialize the library with the Firebase authen and config
  Firebase.begin(&configF, &auth);
 
  // Get the UID
  Serial.println("Getting User ID");
  while ((auth.token.uid) == "") {
    Serial.print('.');
    delay(1000);
  }
  uid = auth.token.uid.c_str();
  Serial.print("User UID: ");
  Serial.println(uid);
 
  // Update database paths
  readingsPath = "/cameras/" + uid + "/readings";
  databasePath = "/cameras/" + uid;
}


void init_SPIFFS(){
  if (!SPIFFS.begin(true)) {
    Serial.println("An Error has occurred while mounting SPIFFS");
    ESP.restart();
  }
  else {
    delay(500);
    Serial.println("SPIFFS mounted successfully");
  }
}

void init_Camera(){
 // OV2640 camera module
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;

  if (psramFound()) {
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }
  // Camera init
  esp_err_t err = esp_camera_init(&config);
  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    ESP.restart();
  } 
}


// Check if photo capture was successful
bool checkPhoto( fs::FS &fs ) {
  File f_pic = fs.open( FILE_PHOTO );
  unsigned int pic_sz = f_pic.size();
  return ( pic_sz > 100 );
}

// Capture Photo and Save it to SPIFFS
void capturePhotoSaveSpiffs( void ) {
  camera_fb_t * fb = NULL; // pointer
  bool ok = 0; // Boolean indicating if the picture has been taken correctly
  do {
    // Take a photo with the camera
    Serial.println("Taking a photo...");

    fb = esp_camera_fb_get();
    if (!fb) {
      Serial.println("Camera capture failed");
      return;
    }
    // Photo file name
    Serial.printf("Picture file name: %s\n", FILE_PHOTO);
    File file = SPIFFS.open(FILE_PHOTO, FILE_WRITE);
    // Insert the data in the photo file
    if (!file) {
      Serial.println("Failed to open file in writing mode");
    }
    else {
      file.write(fb->buf, fb->len); // payload (image), payload length
      Serial.print("The picture has been saved in ");
      Serial.print(FILE_PHOTO);
      Serial.print(" - Size: ");
      Serial.print(file.size());
      Serial.println(" bytes");
    }
    // Close the file
    file.close();
    esp_camera_fb_return(fb);

    // check if file has been correctly saved in SPIFFS
    ok = checkPhoto(SPIFFS);
  } while ( !ok );
}


void setup() {
  pinMode(BUILTIN_RED_LED, OUTPUT);
  digitalWrite(BUILTIN_RED_LED, HIGH);  // LED is off

  Serial.begin(115200);

  init_SPIFFS();
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);  // Turn-off the 'brownout detector'
  init_Camera();  

  dht.begin();
  last_read = 0;
  last_hourly = 0;
  temp_med_hour = 0;
  pres_med_hour = 0;
  current_day_5 = 0;
  current_day_30 = 0;

  init_WiFi();

  // NTP and Firebase require a working WiFi connection
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);  // Initialize the NTP client
  byte hour = get_local_time();
  Serial.println(hour);

  init_Firebase();
}

void loop() {
  unsigned long current_time = millis();
  if (Firebase.ready() && (current_time - last_read > DHT_TIMEOUT || last_read == 0)) {
    capturePhotoSaveSpiffs();

    // Read from the sensor; a reading takes about 250 milliseconds
    float humidity = dht.readHumidity();
    float temperature = dht.readTemperature();  // as Celsius
    if (isnan(humidity) || isnan(temperature)) {
      Serial.println(F("Failed to read from DHT sensor!"));
    }
    else {
      last_read = current_time;
      float hic = dht.computeHeatIndex(temperature, humidity, false);  // in Celsius

      nr_samples++;
      readings_json.set(humidityPath.c_str(), String(humidity));
      readings_json.set(tempPath.c_str(), String(temperature));
      readings_json.set(heatPath.c_str(), String(hic));

      // Serial.print(F("Humidity: "));
      // Serial.print(humidity);
      // Serial.print(F("%  Temperature: "));
      // Serial.print(temperature);
      // Serial.print(F("°C "));
      // Serial.print(F("Heat index: "));
      // Serial.print(hic);
      // Serial.println(F("°C "));

      Serial.printf("Set readings json... %s\n",
                    Firebase.RTDB.setJSON(&fbdo, readingsPath.c_str(), &readings_json) ? "ok" : fbdo.errorReason().c_str());

      // MIME type should be valid to avoid the download problem.
      // The file systems for flash and SD/SDMMC can be changed in FirebaseFS.h.
      
      delay(500);
      if (Firebase.Storage.upload(&fbdo, STORAGE_BUCKET_ID, FILE_PHOTO,
                                  mem_storage_type_flash /* memory storage type, mem_storage_type_flash and mem_storage_type_sd */,
                                  FILE_PHOTO /* path of remote file stored in the bucket */,
                                  "image/jpeg" /* mime type */)) {
        String feed_url = fbdo.downloadURL().c_str();
        Serial.printf("\nDownload URL: %s\n", feed_url);

        live_json.set(live.c_str(), String(feed_url));
        Serial.printf("Set live json... %s\n",
                      Firebase.RTDB.setJSON(&fbdo, databasePath.c_str(), &live_json) ? "ok" : fbdo.errorReason().c_str());
      }
      else {
        Serial.print("Picture not saved to storage because ");
        Serial.println(fbdo.errorReason());
      }

      if (current_time - last_hourly > HOUR_LONG || last_read == 0) {
        // Update the hourly values of the current day
        byte hour = get_local_time();
        Serial.println(hour);

        last_hourly = current_time;
        temp_med_hour /= nr_samples;
        pres_med_hour /= nr_samples;

        String last_5_current_day = databasePath + "/" + last5 + "/" + String(current_day_5);
        String last_5_current_hour = last_5_current_day + "/" + String(hour);
        String last_5_current_picture = "cam1/" + last5 + "/" + String(current_day_5) + "/" + String(hour);

        last5_json.set(tempPath.c_str(), String(temp_med_hour));
        last5_json.set(humidityPath.c_str(), String(pres_med_hour));

        delay(500);
        if (Firebase.Storage.upload(&fbdo, STORAGE_BUCKET_ID, FILE_PHOTO,
                                    mem_storage_type_flash /* memory storage type, mem_storage_type_flash and mem_storage_type_sd */,
                                    last_5_current_picture /* path of remote file stored in the bucket */,
                                    "image/jpeg" /* mime type */)) {
          String feed_url = fbdo.downloadURL().c_str();
          last5_json.set(live.c_str(), String(feed_url));
        }
        else {
          Serial.print("Picture not saved to storage because ");
          Serial.println(fbdo.errorReason());
        }
        Serial.printf("Set last 5 days json... %s\n",
                      Firebase.RTDB.setJSON(&fbdo, last_5_current_hour.c_str(), &last5_json) ? "ok" : fbdo.errorReason().c_str());
        
        temp_med_hour = 0;
        pres_med_hour = 0;
        nr_samples = 0;

        current_day_5++;
        if (current_day_5 == 5) {
          current_day_5 = 0;
        }
        if (hour == 0 || hour == 12 || hour == 5 || hour == 18) {
          // One of the 4 key moments for which pictures are kept
          String last_30_pictures = databasePath + "/" + last30 + "/" + String(current_day_30) + "/" + String(hour);
          String last_30_current_picture = "cam1/" + last30 + "/" + String(current_day_30) + "/" + String(hour);

          delay(500);
          if (Firebase.Storage.upload(&fbdo, STORAGE_BUCKET_ID, FILE_PHOTO,
                                      mem_storage_type_flash /* memory storage type, mem_storage_type_flash and mem_storage_type_sd */,
                                      last_30_current_picture /* path of remote file stored in the bucket */,
                                      "image/jpeg" /* mime type */)) {
            String feed_url = fbdo.downloadURL().c_str();
            last30_json.set(live.c_str(), String(feed_url));
            Serial.printf("Set last 30 days json... %s\n",
                          Firebase.RTDB.setJSON(&fbdo, last_30_pictures.c_str(), &last30_json) ? "ok" : fbdo.errorReason().c_str());
          }
          else {
            Serial.print("Picture not saved to storage because ");
            Serial.println(fbdo.errorReason());
          }
        }
        if (hour == 0) {
          // End of the day
          float day_temp = 0;
          float day_hum = 0;
          String last_5_hour_i, last_5_hour_temp, last_5_hour_hum;

          for (int i = 0; i < 24; i++) {
            last_5_hour_i = last_5_current_day + "/" + String(i);
            last_5_hour_temp = last_5_hour_i + tempPath;
            last_5_hour_hum = last_5_hour_i + humidityPath;

            if (Firebase.RTDB.getFloat(&fbdo, last_5_hour_temp)) {
              if (fbdo.dataType() == "float") {
                day_temp += fbdo.floatData();
              }
            }
            else {
              Serial.println(fbdo.errorReason());
            }

            if (Firebase.RTDB.getFloat(&fbdo, last_5_hour_hum)) {
              if (fbdo.dataType() == "float") {
                day_hum += fbdo.floatData();
              }
            }
            else {
              Serial.println(fbdo.errorReason());
            }
          }
          day_temp /= 24;
          day_hum /= 24;
          String last_30_pictures = databasePath + "/" + last30 + "/" + String(current_day_30) + "/" + String(hour);
          last30_json.set(tempPath.c_str(), String(day_temp));
          last30_json.set(humidityPath.c_str(), String(day_hum));
          Serial.printf("Set last 30 days json... %s\n",
                        Firebase.RTDB.setJSON(&fbdo, last_30_pictures.c_str(), &last30_json) ? "ok" : fbdo.errorReason().c_str());
          
          current_day_30++;
          if (current_day_30 == 30) {
            current_day_30 = 0;
          }          
        }        
      }
    }    
  }
}
